import axios from 'axios';
import bodyParser from 'body-parser';
import { createHmac, randomBytes } from 'crypto';
import { NextFunction, Request, Response } from 'express';
import https from 'https';
import { ApiError } from '../model/apiError';
import { IServiceDescriptor } from '../model/serviceDescriptor';
import { IAuthRequest } from './model/authRequest';
import { ICoronerRequestOption } from './model/authRequestOptions';
import { IServerConfiguration } from './model/serverConfiguration';

/**
 * Identity manager
 */
export class IdentityManager {
  /*
   * Stores opaque objects per authenticated remote coronerd.
   * Available for consumers to use as they see fit.
   */
  public authenticated: { [index: string]: object } = {};

  private logger:
    | { log: (level: string, log: string) => void }
    | undefined = undefined;

  private httpsAgent = this.descr.isOnPremise
    ? new https.Agent({ rejectUnauthorized: false })
    : https.globalAgent;

  constructor(private descr: IServiceDescriptor) {}

  public setLogger(logger: { log: (level: string, msg: string) => void }) {
    if (!logger || !logger.log) {
      throw new TypeError('unusable logger passed in');
    }
    this.logger = logger;
  }

  /*
   * Generate a middleware handler for Express.js, using options to hook the
   * service up.
   *
   * @param opts ICoronerRequestOption - coronerd url callback if desired
   */
  public serviceRequest(opts?: ICoronerRequestOption) {
    /*
     * Handle <prefix>/service requests.  These require action QS.  Other QS
     * are defined per action.
     */
    return [
      bodyParser.json(),
      (request: Request, response: Response, next: NextFunction) => {
        if (!request.body) {
          throw new ApiError('Missing body', 400);
        }
        const action = request.body.action;

        if (!action) {
          throw new ApiError('Missing action parameter', 400);
        }

        switch (action) {
          case 'register':
            const { url, nonce, hmac } = request.body;
            if (!url || !nonce || !hmac) {
              throw new ApiError(
                `missing parameters
              ${!url ? 'url' : ''}
              ${!nonce ? 'nonce' : ''}
              ${!hmac ? 'hmac' : ''}`,
                400,
              );
            }
            if (!this.checkHmac(this.descr.secret, nonce, hmac)) {
              throw new ApiError('Unauthorized', 403);
            }

            /* Track authenticated URLs. */
            if (!this.authenticated[url]) {
              this.authenticated[url] = {};
            }

            if (opts && opts.coronerdCallback) {
              opts.coronerdCallback(url);
            }

            response.json({
              status: 'ok',
            });
            break;
          default:
            throw new ApiError('invalid action', 400);
        }
      },
    ];
  }

  /**
   * Generate a middleware for Express.js, which authenticates a request given
   * its URL and session token, provided by an earlier callback in
   * req.coronerAuth.
   * Upon success, the configuration data will be in req.coronerAuth.cfg.
   */
  public authenticateRequest() {
    return async (request: Request, response: Response, next: NextFunction) => {
      const { url, token } = this.readCoronerAuthInfo(request as IAuthRequest);
      if (!token || !url) {
        next(
          new ApiError(
            `${request.ip}: Missing internal params = token or url`,
            403,
          ),
        );
        return;
      }
      try {
        const serverConfiguration = await this.authenticate(url, token);
        if (!serverConfiguration || !serverConfiguration.user) {
          next(new ApiError('Auth failure', 403));
          return;
        }
        const authRequest = (request as IAuthRequest).coronerAuth;
        if (authRequest) {
          authRequest.cfg = serverConfiguration;
        }
        next();
      } catch (err) {
        next(err);
      }
    };
  }

  /**
   * Use coronerd auth method to validate url and session token provided by user
   * For Express apps, consider using the authenticateRequest middleware instead
   * @param url coronerd url
   * @param token coronerd token
   */
  public async authenticate(
    url: string,
    token: string,
  ): Promise<IServerConfiguration> {
    if (!token || !url) {
      throw new ApiError('missing url/token', 400);
    }
    const prefix = url.endsWith('/') ? '' : '/';
    const result = await axios.get<IServerConfiguration>(
      `${url}${prefix}api/config`,
      {
        httpsAgent: this.httpsAgent,
        headers: {
          'X-Service-Name': this.descr.name,
          'X-Coroner-Token': this.sanitizeHeader(token),
          'X-Service-HMAC': this.generateHmac(this.descr.secret, token),
        },
      },
    );
    if (result.status !== 200) {
      throw new ApiError(
        `Invalid server response code ${result.status}`,
        result.status,
      );
    }

    /* Verify the coronerd knows the service secret. */
    const coronerd_nonce = result.headers['x-service-nonce'];
    const coronerd_hmac = result.headers['x-service-hmac'];
    if (!coronerd_nonce || !coronerd_hmac) {
      throw new ApiError(
        'Missing response parameters - coronerd_nonce/hmac',
        400,
      );
    }
    if (!this.checkHmac(this.descr.secret, coronerd_nonce, coronerd_hmac)) {
      throw new ApiError('Invalid server generated HMAC', 400);
    }
    return result.data as IServerConfiguration;
  }

  /**
   * Login to a coronerd.
   * @param Coronerd base URL to login to
   */
  public async loginCoronerd(
    coronerdUrl: string,
  ): Promise<IServerConfiguration | undefined> {
    if (!coronerdUrl) {
      throw new Error('CoronerdUrl is undefined or empty string');
    }
    const prefix = coronerdUrl.endsWith('/') ? '' : '/';
    const protocol = coronerdUrl.startsWith('http') ? '' : 'http://';
    const requestUrl = `${protocol}${coronerdUrl}${prefix}api/login`;
    const nonce = randomBytes(32).toString('hex');
    const hmac = this.generateHmac(this.descr.secret, nonce);

    try {
      const result = await axios.post<IServerConfiguration>(requestUrl, null, {
        httpsAgent: this.httpsAgent,
        headers: {
          'X-Service-Name': this.descr.name,
          'X-Service-Nonce': nonce,
          'X-Service-HMAC': hmac,
        },
      });
      return result.status === 200 ? result.data : undefined;
    } catch (err) {
      this.log('error', `Cannot login to coronerd. Reason ${err.message}`);
      return undefined;
    }
  }

  /**
   * @deprecated: these methods will be removed in new backtrace-service  update
   * Check if authtoken is a valid token generated by coronerd
   * @param universeUrl url to coronerd universe
   * @param authToken user token
   */
  public async isValidToken(
    universeUrl: string,
    authToken: string,
  ): Promise<boolean> {
    const result = await this.getConfiguration(universeUrl, authToken);
    return result !== undefined;
  }

  /**
   * @deprecated: these methods will be removed in new backtrace-service  update
   * Get Backtrace configuration
   * @param universeUrl url to coronerd universe
   * @param authToken user token
   */
  public async getConfiguration(
    universeUrl: string,
    authToken: string,
  ): Promise<IServerConfiguration | undefined> {
    if (!authToken || !universeUrl) {
      return undefined;
    }
    const prefix = universeUrl.endsWith('/') ? '' : '/';
    const protocol = universeUrl.startsWith('http') ? '' : 'http://';
    const requestUrl = `${protocol}${universeUrl}${prefix}api/config`;
    try {
      const result = await axios.get<IServerConfiguration>(requestUrl, {
        headers: {
          'X-Coroner-Token': authToken,
        },
      });
      return result.status === 200 ? result.data : undefined;
    } catch (err) {
      return undefined;
    }
  }

  private generateHmac(secret: string, nonce: string) {
    if (!secret || !nonce) {
      return '';
    }
    return createHmac('sha256', secret)
      .update(nonce)
      .digest('hex');
  }

  private checkHmac(
    secret: string,
    nonce: string,
    proposedHMac: string,
  ): boolean {
    const hmac = this.generateHmac(secret, nonce);
    if (!hmac) {
      return false;
    }
    return hmac === proposedHMac;
  }

  private sanitizeHeader(value: string) {
    if (!value) {
      return '';
    }
    return value.replace(/[^a-z0-9A-Z?!]/gi, '');
  }

  private log(level: string, msg: string) {
    if (this.logger) {
      this.logger.log(level, msg);
    }
  }

  private readCoronerAuthInfo(request: IAuthRequest) {
    let data = request.coronerAuth;
    if (!data) {
      data = {
        token: this.sanitizeHeader(request.get('X-Coroner-Token') as string),
        url:
          (request.get('X-Coroner-Location') as string) ||
          (request.get('origin') as string),
      };
      if (data.url && !data.url.startsWith('http')) {
        data.url = `http://${data.url}`;
      }
      request.coronerAuth = data;
    }
    return data;
  }
}

// vim:set ts=2 sw=2 et:
