import {
  BacktraceClientOptions,
  getBacktraceClient,
  IBacktraceClientOptions,
  initialize,
} from 'backtrace-node';
import winston, { configure, format, transports } from 'winston';

export class Logger {
  public static getLogger(): Logger {
    if (!Logger._logger) {
      this._logger = new Logger();
      this._logger.warn(`Creating new default logger instance`);
    }

    return this._logger;
  }

  public static initializeBacktrace(
    opts: BacktraceClientOptions | IBacktraceClientOptions,
  ) {
    if (!opts || !opts.endpoint || !opts.token) {
      return;
    }
    initialize(opts);
  }

  private static _logger: Logger;

  // transportLevel definition that Logger class respect:
  // #define SD_EMERG   "<0>"  /* system is unusable */
  // #define SD_ALERT   "<1>"  /* action must be taken immediately */
  // #define SD_CRIT    "<2>"  /* critical conditions */
  // #define SD_ERR     "<3>"  /* error conditions */
  // #define SD_WARNING "<4>"  /* warning conditions */
  // #define SD_NOTICE  "<5>"  /* normal but significant condition */
  // #define SD_INFO    "<6>"  /* informational */
  // #define SD_DEBUG   "<7>"  /* debug-level messages */

  private readonly transportLevel: { [index: string]: number } = {
    error: 3,
    warn: 4,
    info: 5,
    verbose: 6,
    debug: 7,
    silly: 6,
  };

  private transportLevelNumber!: number;
  constructor(private logLevel: string = 'debug') {
    this.transportLevelNumber =
      this.transportLevel[logLevel] || this.transportLevel['debug'];
    configure({
      transports: new transports.Console({
        level: logLevel,
        silent: false,
        format: format.combine(format.colorize(), format.simple()),
      }),
    });
    Logger._logger = this;
  }
  public trace(message: string): void {
    winston.verbose(this.formatScope(message));
  }
  public info(message: string): void {
    winston.info(this.formatScope(message));
  }
  public warn(message: string): void {
    winston.warn(this.formatScope(message));
  }
  public debug(message: string): void {
    winston.debug(this.formatScope(message));
  }
  public error(data: string | Error): void {
    const exception: Error = data instanceof Error ? data : new Error(data);

    const client = getBacktraceClient();
    if (client) {
      client.reportSync(exception);
    }
    winston.error(this.formatScope(exception.message), [exception]);
  }

  private formatScope(msg: string): string {
    // tslint:disable-next-line:max-line-length
    return `<${this.transportLevelNumber}>${this.logLevel}:[${new Date().toLocaleString()}] ${msg}`;
  }
}
