const SamlStrategy = require('passport-saml').Strategy;
const Tenant = require('./models/tenant');
const sendReport = require('./util/sendReport');
const passport = require('passport');

const { service } = require('./util/service-layer');

const config = service.configuration;

function _getConfig(req, cb) {
  Tenant.get(req.params.hostname)
    .then((data) => {
      const tenantConfig = data[0];

      if (tenantConfig && typeof tenantConfig === 'object') {
        let settings = {};

        if (tenantConfig.settings) {
          try {
            settings = JSON.parse(tenantConfig.settings);
          } catch (e) {
            sendReport(e);
          }
        }
        Object.assign(tenantConfig, { settings });

        req.session.tenant = tenantConfig;

        let gateway;

        if (typeof config.gateway === 'object') {
          Object.keys(config.gateway).forEach((g) => {
            if (
              Array.isArray(config.gateway[g]) &&
              config.gateway[g].indexOf(req.params.hostname) >= 0
            ) {
              gateway = g;
            }
          });
        }

        return cb(
          null,
          req.params.hostname,
          Object.assign(
            {
              gateway: gateway ? gateway : req.params.hostname,
              tenant: req.params.hostname,
            },
            tenantConfig
          )
        );
      } else {
        return cb(new Error('Tenant not found'));
      }
    })
    .catch((error) => {
      sendReport(error, { hostname: req.params.hostname });
      cb(error);
    });
}

function _createInstance(options, cb) {
  const instance = new passport.Passport();

  // merge configuration overrides
  const configSamlOptions = Object.assign({}, config.samlOptions);

  // To set up ACS Url there are two approaches to take
  // 1. on premise service - use `callbackUrl` key with an absolute URL ("https://<hostname>/<pathPrefix>/<hostname>")
  // 2. hosted service - use `path` with a relative path ("/<pathPrefix>/<hostname>")

  // set up for legacy backtrace-saml clients who already have an ACS Url
  // that takes the form https://<saml host>/saml/<host>
  // which requires a saml.conf setting of:
  // "samlOptions": { "pathPrefix": "/saml" }
  // new ACS url will take the form of https://<host>/api/sso/saml/<saml>
  const pathPrefix = configSamlOptions.pathPrefix
    ? configSamlOptions.pathPrefix
    : config.pathPrefix + '/saml';

  // TODO remove this check and just have path unless callbackUrl is in the Record...
  if (service.identityManager.isOnPremise) {
    // is there a callbackUrl specified in the saml.conf file if not, create one for our options
    // NOTE: there is now a callbackUrl column so once that is utilized by the web console's create form, then
    // this will be overridden
    if (!configSamlOptions.callbackUrl) {
      configSamlOptions.callbackUrl = `https://${options.gateway}${pathPrefix}/${options.gateway}`;
    }
  } else {
    // This is for hosted solutions where can control the hostname with the descriptor
    configSamlOptions.path = `${pathPrefix}/${options.gateway}`;
  }

  const samlOptions = Object.assign(
    {
      disableRequestedAuthnContext: true,
      entryPoint: options.entryPoint,
      issuer: options.issuer,
    },
    configSamlOptions
  );

  // options to be added to the saml settings from the tenant record
  const optional = [
    'cert',
    'privateCert',
    'decryptionPvk',
    'callbackUrl',
    'signatureAlgorithm',
  ];

  // TODO: After persistance layer is added - these files will just be strings
  optional.forEach((key) => {
    if (options[key]) {
      samlOptions[key] = options[key];
    }
  });

  instance.use(
    'saml',
    new SamlStrategy(samlOptions, (profile, cb) => {
      cb(null, profile);
    })
  );

  instance.serializeUser(function serializeUserCallback(user, cb) {
    user.tenant = options.tenant;
    cb(null, JSON.stringify(user));
  });

  instance.deserializeUser(function deserializeUserCallback(id, cb) {
    cb(null, JSON.parse(id));
  });

  cb(null, instance);
}

module.exports = {
  getConfig: _getConfig,
  createInstance: _createInstance,
};
