const Tenant = require('../models/tenant');
const sendReport = require('./sendReport');
const { service } = require('./service-layer');

const config = service.configuration;

// Function for transforming the tenantConfig object from our database into
// somethign more useable for us
function unwrapConfiguration(req, tenantConfig, fn) {
  // parse settings stringified JSON
  if (tenantConfig.settings) {
    let settings = {};
    try {
      settings = JSON.parse(tenantConfig.settings);
    } catch (e) {
      fn(e);
    }
    Object.assign(tenantConfig, { settings });
  }

  // get gateway in event multiple universes are using the same IdP settings
  if (config.gateway && typeof config.gateway === 'object') {
    let gateway = undefined;
    Object.keys(config.gateway).forEach((g) => {
      if (
        Array.isArray(config.gateway[g]) &&
        config.gateway[g].indexOf(req.params.hostname) >= 0
      ) {
        gateway = g;
      }
    });
    Object.assign(tenantConfig, {
      gateway: gateway ? gateway : req.params.hostname,
      tenant: req.params.hostname,
    });
  }

  fn(null, tenantConfig);
}

// Function to trim certificate strings to single line strings without certificate
// declaration lines (these are added and handled by passport-saml)
function processCertificate(cert) {
  let processedCertificate = cert;
  if (processedCertificate.includes('-----BEGIN PRIVATE KEY-----')) {
    processedCertificate = processedCertificate
      .replace('-----BEGIN PRIVATE KEY-----', '')
      .replace('-----END PRIVATE KEY-----', '');
  } else if (processedCertificate.includes('-----BEGIN RSA PRIVATE KEY-----')) {
    processedCertificate = processedCertificate
      .replace('-----BEGIN RSA PRIVATE KEY-----', '')
      .replace('-----END RSA PRIVATE KEY-----', '');
  } else if (processedCertificate.includes('-----BEGIN CERTIFICATE-----')) {
    processedCertificate = processedCertificate
      .replace('-----BEGIN CERTIFICATE-----', '')
      .replace('-----END CERTIFICATE-----', '');
  }
  try {
    return processedCertificate.trim().split('\n').join('');
  } catch (e) {
    sendReport(e);
    return cert;
  }
}

function getSamlOptionsFromTenant(tenantConfig) {
  // merge configuration overrides
  const configSamlOptions = Object.assign({}, config.samlOptions);

  // To set up ACS Url there are two approaches to take
  // 1. on premise service - use `callbackUrl` key with an absolute URL ("https://<hostname>/<pathPrefix>/<hostname>")
  // 2. hosted service - use `path` with a relative path ("/<pathPrefix>/<hostname>")

  // set up for legacy backtrace-saml clients who already have an ACS Url
  // that takes the form https://<saml host>/saml/<host>
  // which requires a saml.conf setting of:
  // "samlOptions": { "pathPrefix": "/saml" }
  // new ACS url will take the form of https://<host>/api/sso/saml/<saml>
  const pathPrefix = configSamlOptions.pathPrefix
    ? configSamlOptions.pathPrefix
    : config.pathPrefix + '/saml';

  // TODO remove this check and just have path unless callbackUrl is in the Record...
  if (service.identityManager.isOnPremise) {
    // is there a callbackUrl specified in the saml.conf file if not, create one for our options
    // NOTE: there is now a callbackUrl column so once that is utilized by the web console's create form, then
    // this will be overridden
    if (!configSamlOptions.callbackUrl) {
      configSamlOptions.callbackUrl = `https://${tenantConfig.gateway}${pathPrefix}/${tenantConfig.gateway}`;
    }
  } else {
    // This is for hosted solutions where can control the hostname with the descriptor
    configSamlOptions.path = `${pathPrefix}/${tenantConfig.gateway}`;
  }

  const samlOptions = Object.assign(
    {
      disableRequestedAuthnContext: true,
      entryPoint: tenantConfig.entryPoint,
      issuer: tenantConfig.issuer,
    },
    configSamlOptions
  );

  // options to be added to the saml settings from the tenant record
  const optional = ['callbackUrl', 'signatureAlgorithm'];

  const optionalCerts = ['cert', 'privateCert', 'decryptionPvk'];

  // Add optional params
  optional.forEach((key) => {
    if (tenantConfig[key]) {
      samlOptions[key] = tenantConfig[key];
    }
  });

  // Add optional certs / keys AND process them
  optionalCerts.forEach((key) => {
    if (tenantConfig[key]) {
      samlOptions[key] = processCertificate(tenantConfig[key]);
    }
  });

  return samlOptions;
}

module.exports = function getSamlOptions(req, done) {
  Tenant.get(req.params.hostname)
    .then((data) => {
      const tenantConfig = data[0];

      if (!tenantConfig || typeof tenantConfig !== 'object') {
        return done(new Error('no tenant found'));
      }

      unwrapConfiguration(req, tenantConfig, function (err, newTenantConfig) {
        if (err) {
          return done(err);
        }

        // Stuff tenantConfig into session
        req.session.tenant = newTenantConfig;

        const samlOptions = getSamlOptionsFromTenant(newTenantConfig);

        return done(null, samlOptions);
      });
    })
    .catch(done);
};
